/*
 * Plugin that writes its input to a wav file.
 *
 * Eric Fikus <fikus8c0@erau.edu>
 */

#include <stdlib.h>
#include <plugin.h>
#include <string.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <fcntl.h>
#include <gtk/gtk.h>
#include <unistd.h>

#define RATE 44100

typedef struct {
  guint written;
  gboolean enabled;
  gchar *filename;
  GtkWidget *window;
} pluginData;

typedef enum
{
  FMT_U8, FMT_S8, FMT_U16_LE, FMT_U16_BE, FMT_U16_NE, FMT_S16_LE, FMT_S16_BE, FMT_S16_NE
}
AFormat;

struct wavhead
{
  guint32 main_chunk;
  guint32 length;
  guint32 chunk_type;
  guint32 sub_chunk;
  guint32 sc_len;
  guint16 format;
  guint16 modus;
  guint32 sample_fq;
  guint32 byte_p_sec;
  guint16 byte_p_spl;
  guint16 bit_p_spl;
  guint32 data_chunk;
  guint32 data_length;
};

static AFormat getfmt(void);

static FILE *output_file;
static struct wavhead header;
static AFormat afmt;

void *init(void)
{
  pluginData *data;
  gchar *default_filename;
  
  default_filename = calloc(256, sizeof(char));
  g_snprintf(default_filename, 256, "%s%s", g_get_home_dir(), "/stompbox-output.wav");
  
  data  = malloc(sizeof(pluginData));
  data->written = 0;
  data->filename = calloc(256, sizeof(char));
  strcpy(data->filename, default_filename);
  free(default_filename);
  data->enabled = TRUE;
  data->window = NULL;

  afmt = getfmt();

  return data;
}

static void cleanup(void *data)
{ 
  pluginData *temp = data;

  if (output_file) {
    header.length = GUINT32_TO_LE(temp->written + sizeof (struct wavhead) - 8);
    
    header.data_length = GUINT32_TO_LE(temp->written);
    fseek(output_file, 0, SEEK_SET);
    fwrite(&header, sizeof(struct wavhead), 1, output_file);
    
    fclose(output_file);
    temp->written = 0;
  }
  
  output_file = NULL;

  free(temp->filename);
  free(data);
  
}

static gint window_destroy(GtkWidget *window, gpointer data)
{
   pluginData *plug = data;
   gtk_widget_destroy(window);
   plug->window = NULL;
   return TRUE;
}

static gint filename_changed(GtkWidget *widget, gpointer data)
{
  pluginData *temp = data;
  strcpy(temp->filename, gtk_entry_get_text(GTK_ENTRY(widget)));

  temp->written = 0;

  return TRUE;
}

static gint enabled_toggled(GtkWidget *widget, gpointer data)
{
  pluginData *temp = data;
  temp->enabled = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

  return TRUE;
}

static void configure(void *data)
{
  GtkWidget *table;
  GtkWidget *filename_label;
  GtkWidget *filename_entry;
  GtkWidget *enabled_label;
  GtkWidget *enabled_button;
  pluginData *temp = data;

  if (temp->window == NULL) {
    temp->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(temp->window), "Waveout");
    gtk_widget_set_usize(GTK_WIDGET(temp->window), 300, 50);
    gtk_signal_connect(GTK_OBJECT(temp->window), "destroy", GTK_SIGNAL_FUNC(window_destroy), data);
    
    table = gtk_table_new(2, 2, FALSE);
    gtk_container_add(GTK_CONTAINER(temp->window), table);

    filename_label = gtk_label_new("Output filename");
    filename_entry = gtk_entry_new();
    enabled_label = gtk_label_new("Enabled");
    enabled_button = gtk_check_button_new();
    gtk_entry_set_text(GTK_ENTRY(filename_entry), temp->filename);
    gtk_signal_connect(GTK_OBJECT(filename_entry), "changed", GTK_SIGNAL_FUNC(filename_changed), data);
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(enabled_button), temp->enabled);
    gtk_signal_connect(GTK_OBJECT(enabled_button), "toggled", GTK_SIGNAL_FUNC(enabled_toggled), data);
    gtk_table_attach_defaults(GTK_TABLE(table), filename_label, 0, 1, 0, 1);
    gtk_table_attach_defaults(GTK_TABLE(table), filename_entry, 1, 2, 0, 1);
    gtk_table_attach_defaults(GTK_TABLE(table), enabled_label, 0, 1, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(table), enabled_button, 1, 2, 1, 2);
    gtk_widget_show_all(GTK_WIDGET(temp->window));

  }
    
}

static AFormat getfmt(void)
{
  gint format;
  gint audio_fd;

  if ((audio_fd = open("/dev/dsp", O_RDONLY)) == -1) {
    perror("waveout.c open()");
    return -1;
  }

  if (ioctl(audio_fd, SNDCTL_DSP_GETFMTS, &format) < 0) {
    perror("waveout.c SNDCTL_DSP_GETFMTS");
    return -1;
  }

  close(audio_fd);

  if (format & AFMT_S16_NE)
    return FMT_S16_NE;
  if (format & AFMT_S16_BE)
    return FMT_S16_BE;
  if (format & AFMT_S16_LE)
    return AFMT_S16_LE;
  /*
  if (format & AFMT_U16_NE)
    return FMT_U16_NE;
  */
  if (format & AFMT_U16_BE)
    return FMT_U16_BE;
  if (format & AFMT_U16_LE)
    return FMT_U16_LE;
  if (format & AFMT_S8)
    return FMT_S8;

  /* if nothing else, guess unsinged 8-bit format */
  return FMT_U8;
}

static void convert_buffer(gpointer buffer, gint length)
{
  gint i;

  if (afmt == FMT_S8) {
      guint8 *ptr1 = buffer;
      gint8 *ptr2 = buffer;
      
      for (i = 0; i < length; i++)
	*(ptr1++) = *(ptr2++) ^ 128;
  }
  
  if (afmt == FMT_S16_BE) {
      gint16 *ptr = buffer;
      
      for (i = 0; i < length >> 1; i++, ptr++)
	*ptr = GUINT16_SWAP_LE_BE(*ptr);
  }

  if (afmt == FMT_S16_NE) {
      gint16 *ptr = buffer;
      
      for (i = 0; i < length >> 1; i++, ptr++)
	*ptr = GINT16_TO_LE(*ptr);
  }

  if (afmt == FMT_U16_BE) {
      gint16 *ptr1 = buffer;
      guint16 *ptr2 = buffer;
      
      for (i = 0; i < length >> 1; i++, ptr2++)
	*(ptr1++) = GINT16_TO_LE(GUINT16_FROM_BE(*ptr2) ^ 32768);
  }

  if (afmt == FMT_U16_LE) {
      gint16 *ptr1 = buffer;
      guint16 *ptr2 = buffer;
      
      for (i = 0; i < length >> 1; i++, ptr2++)
	*(ptr1++) = GINT16_TO_LE(GUINT16_FROM_LE(*ptr2) ^ 32768);
  }
  
  if (afmt == FMT_U16_NE) {
    gint16 *ptr1 = buffer;
    guint16 *ptr2 = buffer;
    
    for (i = 0; i < length >> 1; i++, ptr2++)
      *(ptr1++) = GINT16_TO_LE((*ptr2) ^ 32768);
  }
  
}

static int process(void *data, short int **smpdata, int size, int sampf)
{
  pluginData *temp = data;
  gshort *channel;
  gint i;

  if (!temp->enabled)
    return 1;

  if (temp->written == 0) {

    memcpy(&header.main_chunk, "RIFF", 4);
    header.length = GUINT32_TO_LE(0);
    memcpy(&header.chunk_type, "WAVE", 4);
    memcpy(&header.sub_chunk, "fmt ", 4);
    header.sc_len = GUINT32_TO_LE(16);
    header.format = GUINT16_TO_LE(1);
    header.modus = GUINT16_TO_LE(1);
    header.sample_fq = GUINT32_TO_LE(RATE);
    if (afmt == FMT_U8 || afmt == FMT_S8)
      header.bit_p_spl = GUINT16_TO_LE(8);
    else
      header.bit_p_spl = GUINT16_TO_LE(16);
    header.byte_p_sec = GUINT32_TO_LE(RATE * header.modus * (GUINT16_FROM_LE(header.bit_p_spl) / 8));
    header.byte_p_spl = GUINT16_TO_LE((GUINT16_FROM_LE(header.bit_p_spl) / (8 / 2)));
    memcpy(&header.data_chunk, "data", 4);
    header.data_length = GUINT32_TO_LE(0);
    
    if ((output_file = fopen(temp->filename, "wb")) == NULL) {
      perror("Error opening output file");
      return 0;
    }
    
    fwrite(&header, sizeof (struct wavhead), 1, output_file);
    
  }

#ifdef WORDS_BIGENDIAN
  if (afmt == FMT_S8 || afmt == FMT_S16_BE || afmt == FMT_S16_NE || afmt == FMT_U16_LE || afmt == FMT_U16_BE || afmt == FMT_U16_NE)
#else
    if (afmt == FMT_S8 || afmt == FMT_S16_BE || afmt == FMT_U16_LE || afmt == FMT_U16_BE || afmt == FMT_U16_NE)
#endif
      convert_buffer(*smpdata, size);

  channel = smpdata[0];

  for (i = 0; i < size; ++i)
    temp->written += fwrite(&channel[i], sizeof channel[i], 1, output_file);

  header.length = GUINT32_TO_LE(temp->written + sizeof (struct wavhead) - 8);

  header.data_length = GUINT32_TO_LE(temp->written);
  fseek(output_file, 0, SEEK_SET);
  fwrite(&header, sizeof (struct wavhead), 1, output_file);
  fseek(output_file, temp->written, SEEK_SET);
		
  return 1;
}

static char *get_state(void *data)
{
  pluginData *temp = data;
  gchar *state_string;
  state_string = calloc(256, sizeof(char));
  /* always writes 0 for bytes written, therefore will try to overwrite instead of append to existing file */
  g_snprintf(state_string, 256, "%d %d %s", 0, temp->enabled, temp->filename);
  return state_string;
}

static int set_state(void *data, char *state_string)
{
  pluginData *temp = data;
  sscanf(state_string,"%d %d %s", &(temp->written), &(temp->enabled), temp->filename);
  return 1;
}

Plugin waveout_plugin_data =
  {
    "Waveout",
    cleanup,
    init,
    configure,
    process,
    get_state,
    set_state,
    1,
    1
  };



Plugin *get_plugin_info(void)
{
  return &waveout_plugin_data;
}
